@@ -8,7 +8,7 @@ package Tickit::Term;
 use strict;
 use warnings;
 
-our $VERSION = '0.56';
+our $VERSION = '0.57';
 
 use Carp;
 
@@ -178,7 +178,7 @@ Flushes the output buffer to the terminal
 
 Installs a new event handler to watch for the event specified by C<$ev>,
 invoking the C<$code> reference when it occurs. C<$code> will be invoked with
-the given the terminal, the event name, a C<HASH> reference containing event
+the given terminal, the event name, a C<HASH> reference containing event
 arguments, and the C<$data> value. It returns an ID value that may be used to
 remove the handler by calling C<unbind_event_id>.
 
@@ -209,21 +209,13 @@ The hash will contain C<lines> and C<cols> as integers.
 
 =back
 
-This method is aliased as C<bind_event_default> for forward compatibility to
-account for L</bind_event_with_flags> being renamed to simply C<bind_event>.
+=head2 bind_event (with flags)
 
-=cut
+   $id = $term->bind_event( $ev, $flags, $code, $data )
 
-# Eventually throw a deprecation warning
-*bind_event = \&bind_event_default;
-
-=head2 bind_event_with_flags
-
-   $id = $term->bind_event_with_flags( $ev, $flags, $code, $data )
-
-Similar to L</bind_event> but passes the C<$flags> argument too. This should
-be zero to apply default semantics, or a bitmask of one or more of the
-following constants:
+The C<$code> argument may optionally be preceeded by an integer of flag
+values. This should be zero to apply default semantics, or a bitmask of one or
+more of the following constants:
 
 =over 4
 
@@ -233,6 +225,9 @@ Inserts this event handler first in the chain, before any existing ones.
 
 =back
 
+(This behaviour used to only be available via a separately named method called
+C<bind_event_with_flags> but that name will shortly be removed.)
+
 =head2 unbind_event_id
 
    $term->unbind_event_id( $id )
@@ -241,6 +236,15 @@ Removes an event handler that returned the given C<$id> value.
 
 =cut
 
+sub bind_event
+{
+   my $self = shift;
+   my $ev = shift;
+   my ( $flags, $code, $data ) = ( ref $_[0] ) ? ( 0, @_ ) : @_;
+
+   $self->bind_event_with_flags( $ev, $flags, $code, $data );
+}
+
 sub set_on_resize
 {
    my $self = shift;
