@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use 5.010; # //
 
-our $VERSION = '0.56';
+our $VERSION = '0.57';
 
 use Carp;
 
@@ -222,11 +222,7 @@ sub _close
 {
    my $self = shift;
 
-   $self->set_on_geom_changed( undef );
-   $self->set_on_key( undef );
-   $self->set_on_mouse( undef );
-   $self->set_on_expose( undef );
-   $self->set_on_focus( undef );
+   $self->unbind_event_id( $_ ) for values %{ $self->{event_ids} };
 
    defined and $_->_close for @{ $self->{child_windows} };
    undef $self->{child_windows};
@@ -407,6 +403,88 @@ sub make_popup
    return $popup;
 }
 
+sub _fire_event
+{
+   my $self = shift;
+   my ( $ev, $info ) = @_;
+
+   foreach my $hook ( @{ $self->{eventhooks} } ) {
+      my ( $hookev, $code, $data ) = @$hook;
+      next unless $ev eq $hookev;
+
+      $code->( $self, $ev, $info, $data ) and return 1;
+   }
+
+   return 0;
+}
+
+=head2 bind_event
+
+   $id = $win->bind_event( $ev, $code, $data )
+
+Installs a new event handler to watch for the event specified by C<$ev>,
+invoking the C<$code> reference when it occurs. C<$code> will be invoked with
+the given window, the event name, an event information object, and the
+C<$data> value it was installed with. C<bind_event> returns an ID value that
+may be used to remove the handler by calling C<unbind_event_id>.
+
+ $ret = $code->( $win, $ev, $info, $data )
+
+This method is preferred over the legacy C<set_on_EVENT> handler methods.
+
+=head2 bind_event (with flags)
+
+   $id = $win->bind_event( $ev, $flags, $code, $data )
+
+The C<$code> argument may optionally be preceeded by an integer of flag
+values. This should be zero to apply default semantics, or a bitmask of
+constants. The constants are documented in
+L<Tickit::Term/bind_event (with flags)>.
+
+=cut
+
+sub bind_event
+{
+   my $self = shift;
+   my $ev = shift;
+   my ( $flags, $code, $data ) = ( ref $_[0] ) ? ( 0, @_ ) : @_;
+
+   my $hooks = $self->{eventhooks} //= [];
+   my $hook = [ $ev, $code, $data ];
+
+   if( $flags & Tickit::Term::BIND_FIRST ) {
+      unshift @$hooks, $hook;
+   }
+   else {
+      push @$hooks, $hook;
+   }
+
+   return 0 + $hook;
+}
+
+=head2 unbind_event_id
+
+   $win->unbind_event_id( $id )
+
+Removes an event handler that returned the given C<$id> value.
+
+=cut
+
+sub unbind_event_id
+{
+   my $self = shift;
+   my ( $id ) = @_;
+
+   my $hooks = $self->{eventhooks};
+
+   foreach my $idx ( 0 .. $#$hooks ) {
+      next unless $hooks->[$idx] == $id;
+
+      splice @$hooks, $idx, 1, ();
+      last;
+   }
+}
+
 =head2 raise
 
    $win->raise
@@ -702,7 +780,7 @@ sub change_geometry
       $self->{top} = $top;
       $self->{left} = $left;
 
-      $self->{on_geom_changed}->( $self ) if $self->{on_geom_changed};
+      $self->_fire_event( geomchange => );
    }
 }
 
@@ -715,12 +793,20 @@ i.e. whenever its geometry changes.
 
  $on_geom_changed->( $win )
 
+This method is now deprecated in favour of L</bind_event>.
+
 =cut
 
 sub set_on_geom_changed
 {
    my $self = shift;
-   ( $self->{on_geom_changed} ) = @_;
+   my ( $code ) = @_;
+
+   $self->unbind_event_id( delete $self->{event_ids}{geomchange} ) if $self->{event_ids}{geomchange};
+   $self->{event_ids}{geomchange} = $self->bind_event( geomchange => sub {
+      my ( $self ) = @_;
+      $code->( $self );
+   } );
 }
 
 =head2 set_on_key
@@ -761,6 +847,8 @@ special action.
 
 The C<with_ev>-prefixed form is accepted for backward compatibility.
 
+This method is now deprecated in favour of L</bind_event>.
+
 =cut
 
 sub set_on_key
@@ -771,7 +859,13 @@ sub set_on_key
       shift;
    }
 
-   ( $self->{on_key} ) = @_;
+   my ( $code ) = @_;
+
+   $self->unbind_event_id( delete $self->{event_ids}{key} ) if $self->{event_ids}{key};
+   $self->{event_ids}{key} = $self->bind_event( key => sub {
+      my ( $self, undef, $info ) = @_;
+      $code->( $self, $info );
+   } ) if $code;
 }
 
 sub _handle_key
@@ -793,9 +887,8 @@ sub _handle_key
       $focused_child->_handle_key( $args ) and return 1;
    }
 
-   if( my $on_key = $self->{on_key} ) {
-      $on_key->( $self, Tickit::Window::Event->new( %$args ) ) and return 1;
-   }
+   my $event = Tickit::Window::Event->new( %$args );
+   $self->_fire_event( key => $event ) and return 1;
 
    if( $children ) {
       foreach my $child ( @$children ) {
@@ -881,6 +974,8 @@ started the operation.
 
 The C<with_ev>-prefixed form is accepted for backward compatibility.
 
+This method is now deprecated in favour of L</bind_event>.
+
 =cut
 
 sub set_on_mouse
@@ -891,7 +986,13 @@ sub set_on_mouse
       shift;
    }
 
-   ( $self->{on_mouse} ) = @_;
+   my ( $code ) = @_;
+
+   $self->unbind_event_id( delete $self->{event_ids}{mouse} ) if $self->{event_ids}{mouse};
+   $self->{event_ids}{mouse} = $self->bind_event( mouse => sub {
+      my ( $self, undef, $info ) = @_;
+      $code->( $self, $info );
+   } ) if $code;
 }
 
 sub _handle_mouse
@@ -927,9 +1028,8 @@ sub _handle_mouse
       }
    }
 
-   if( my $on_mouse = $self->{on_mouse} ) {
-      $on_mouse->( $self, Tickit::Window::Event->new( %$args ) ) and return $self;
-   }
+   my $event = Tickit::Window::Event->new( %$args );
+   $self->_fire_event( mouse => $event ) and return $self;
 
    return 0;
 }
@@ -956,6 +1056,8 @@ The C<with_rb>-prefixed form is accepted for backward compatibility.
 If any child windows overlap the region, these will be exposed first, before
 the containing window.
 
+This method is now deprecated in favour of L</bind_event>.
+
 =cut
 
 sub set_on_expose
@@ -966,7 +1068,13 @@ sub set_on_expose
       shift;
    }
 
-   ( $self->{on_expose} ) = @_;
+   my ( $code ) = @_;
+
+   $self->unbind_event_id( delete $self->{event_ids}{expose} ) if $self->{event_ids}{expose};
+   $self->{event_ids}{expose} = $self->bind_event( expose => sub {
+      my ( $self, undef, $info ) = @_;
+      $code->( $self, $info->rb, $info->rect );
+   } ) if $code;
 }
 
 our $INDENT = "";
@@ -996,13 +1104,11 @@ sub _do_expose
       $rb->mask( $win->rect );
    }
 
-   if( my $on_expose = $self->{on_expose} ) {
-      $rb->save;
+   $rb->save;
 
-      $on_expose->( $self, $rb, $rect );
+   $self->_fire_event( expose => Tickit::Window::Event->new( rb => $rb, rect => $rect ) );
 
-      $rb->restore;
-   }
+   $rb->restore;
 }
 
 =head2 expose
@@ -1077,12 +1183,20 @@ C<focus_child_notify> enabled will be informed first, from the outermost
 inwards, before the window itself. When one loses focus, it is notified
 first, and then its parents from the innermost outwards.
 
+This method is now deprecated in favour of L</bind_event>.
+
 =cut
 
 sub set_on_focus
 {
    my $self = shift;
-   ( $self->{on_focus} ) = @_;
+   my ( $code ) = @_;
+
+   $self->unbind_event_id( delete $self->{event_ids}{focus} ) if $self->{event_ids}{focus};
+   $self->{event_ids}{focus} = $self->bind_event( focus => sub {
+      my ( $self, undef, $info ) = @_;
+      $code->( $self, $info->type, $info->win );
+   } ) if $code;
 }
 
 =head2 set_focus_child_notify
@@ -1102,20 +1216,6 @@ sub set_focus_child_notify
    ( $self->{focus_child_notify} ) = @_;
 }
 
-=head2 set_expose_after_scroll
-
-   $win->set_expose_after_scroll( ... )
-
-This method is now deprecated; the setting is always enabled and can no longer
-be disabled. This method throws an exception if invoked.
-
-=cut
-
-sub set_expose_after_scroll
-{
-   croak "expose_after_scroll is deprecated";
-}
-
 =head2 top
 
 =head2 bottom
@@ -1741,10 +1841,10 @@ sub _focus_gained
 
    if( !$child ) {
       $self->{focused} = 1;
-      $self->{on_focus}->( $self, 1 ) if $self->{on_focus};
+      $self->_fire_event( focus => Tickit::Window::Event->new( type => 1, win => undef ) );
    }
    elsif( $self->{focus_child_notify} ) {
-      $self->{on_focus}->( $self, 1, $child ) if $self->{on_focus};
+      $self->_fire_event( focus => Tickit::Window::Event->new( type => 1, win => $child ) );
    }
 
    $self->{focused_child} = $child;
@@ -1757,12 +1857,12 @@ sub _focus_lost
 
    if( my $focused_child = $self->{focused_child} ) {
       $focused_child->_focus_lost;
-      $self->{on_focus}->( $self, 0, $focused_child ) if $self->{on_focus} && $self->{focus_child_notify};
+      $self->_fire_event( focus => Tickit::Window::Event->new( type => 0, win => $focused_child ) );
    }
 
    if( $self->{focused} ) {
       undef $self->{focused};
-      $self->{on_focus}->( $self, 0 ) if $self->{on_focus};
+      $self->_fire_event( focus => Tickit::Window::Event->new( type => 0, win => undef ) );
    }
 }
 
@@ -1899,15 +1999,15 @@ package # hide from indexer
 
 =head1 EVENT STRUCTURE
 
-Objects in this class are passed to C<on_key> and C<on_mouse> handlers, to
-give details of the keyboard or mouse event that happened.
+Objects in this class are passed to C<bind_event> handlers, to give details of
+the event that happened.
 
-=head2 $ev->type
+=head2 $info->type
 
 A string giving the event type name. The valid values here will differ
-depending on whether it is a keyboard or mouse event.
+depending on whether it is a keyboard, mouse or focus event.
 
-=head2 $ev->str
+=head2 $info->str
 
 A string representing the base key and its modifiers. Modifiers are given as
 an optional prefix; any of C<M-> for Alt (Meta), C<C-> for Control, or C<S->
@@ -1915,22 +2015,34 @@ for Shift (in that order), followed by the base key. For UTF-8 text this will
 be a single character. For non-Unicode special keys this will be the name of
 the key.
 
-=head2 $ev->line
+=head2 $info->line
 
-=head2 $ev->col
+=head2 $info->col
 
 The position for mouse events, 0-based.
 
-=head2 $ev->button
+=head2 $info->button
 
 The button number for non-wheel mouse events, or the mouse wheel direction as
 a string (one fo C<up> or C<down>). Note that for C<release> type events, the
 mouse button may not be reliable; not all terminals can report it.
 
-=head2 $ev->mod
+=head2 $info->mod
 
 A bitmask of modifier state. Valid for both keyboard and mouse events.
 
+=head2 $info->rb
+
+The window's L<Tickit::RenderBuffer> instance for expose events.
+
+=head2 $info->rect
+
+The exposed L<Tickit::Rect> for expose events.
+
+=head2 $info->win
+
+The child window, if any, for focus events.
+
 =cut
 
 use Carp;
@@ -1941,16 +2053,16 @@ sub new
    bless { @_ }, $class;
 }
 
-foreach my $key (qw( type str mod button line col )) {
+foreach my $key (qw( type str mod button line col rect rb win )) {
    no strict 'refs';
    *$key = sub { exists $_[0]{$key} ? $_[0]{$key} : croak "Event has no '$key' field" }
 }
 
-=head2 $ev->mod_is_alt
+=head2 $info->mod_is_alt
 
-=head2 $ev->mod_is_ctrl
+=head2 $info->mod_is_ctrl
 
-=head2 $ev->mod_is_shift
+=head2 $info->mod_is_shift
 
 Convenient shortcuts to tests on the C<mod> bitmask to test if each of the
 modifiers is set.
